#include <DeltaInverseKinematics.h>
#include <Arduino.h>
#include <VarSpeedServo.h>
// #include <SoftwareSerial.h>

#define START_CMD "start" // serial command to start delta sequence
#define FAILED_CMD "failed"
#define TEST_CMD "test"
#define RELAY_CMD "relay"
#define MAX_INACTIVE_TIME 30000 // 30s
#define GODOWN_STEP 10


#define SENSOR A2 // Sharp IR GP2Y0A41SK0F (4-30cm, analog)
#define RELAY_PIN 11
// const int pinMotorSpeed = 6 ;
// const int pinMotorDirection = 11 ;

VarSpeedServo servo3;
VarSpeedServo servo2;
VarSpeedServo servo1;

double angleB1;
double angleB2;
double angleB3;

int positionZ = -150;

int speedServo = 20;
unsigned long previousMillis = 0;
unsigned long startMillis = 0;

// serial Input
char val;
int intval = 0;
char rx_buf[16];
int i_rx=0;

// SoftwareSerial Serial(TX_PIN, RX_PIN); // RX, TX

//DeltaInverseKinematics(double *angleB1,double *angleB2,double *angleB3,double ArmLength,double RodLength,double BassTri,double PlatformTri)
DeltaInverseKinematics IPK(&angleB1,&angleB2,&angleB3,100,220,280,70);

void startDelta();
void updatePosition(double angle1, double angle2, double angle3);
void updateIPK(double X, double Y, double Z, long waitTime);
void magnet(bool state);
void badCode();
void robotDance();
void checkInactive();
int getDistance();
void goDown(int startFromZ, bool debug = false);

void setup()
{
  servo1.attach(2);
  servo2.attach(3);
  servo3.attach(4);

  // pinMode(pinMotorSpeed,OUTPUT);
  // pinMode(pinMotorDirection,OUTPUT);
  //
  // digitalWrite(pinMotorDirection,LOW);
  pinMode(RELAY_PIN,OUTPUT);
  digitalWrite(RELAY_PIN,LOW);

  Serial.begin(9600);
  // robotDance(); // init
}

void updatePosition(double angle1, double angle2, double angle3)
{
  servo1.write(angle1,speedServo);
  servo2.write(angle2,speedServo);
  servo3.write(angle3,speedServo);

  servo1.wait();
  servo2.wait();
  servo3.wait();
}

void updateIPK(double X, double Y, double Z, long waitTime)
{
  IPK.set(X,Y,Z);
  updatePosition(angleB1* 180 / 3.14, angleB2* 180 / 3.14, angleB3* 180 / 3.14);
  delay(waitTime);
}


void magnet(bool state)
{
  // digitalWrite(pinMotorSpeed,state);
  if(state == true){
    digitalWrite(RELAY_PIN,HIGH);
  } else {
    digitalWrite(RELAY_PIN,LOW);
  }
}

void startDelta()
{
  previousMillis = millis();
  updateIPK(-20,-20,-150,500);    // Position High
  magnet(true);                   // Activate Magnet
  // updateIPK(-20,-20,-300,2000);   // Position Low
  updateIPK(-20,-20,-200,100);
  goDown(-200); // detect distance
  // if(positionZ > -250) goDown(positionZ);

  updateIPK(-20,-20,-150,1000);   // Position High
  updatePosition(120,60,130); // Position Right
  delay(1000);
  magnet(false);                  // De-activate Magnet
  delay(500);
  previousMillis = millis();
}

void badCode()
{
  previousMillis = millis();
  updateIPK(-20,-20,-150,500);   // Position High
  // updatePosition(60,100,130); // Position Left
  // updatePosition(-40,100,130); // Position Right
  updatePosition(60,100,130); // Position Left
  updatePosition(100,60,130); // Position Right

  updateIPK(-20,-20,-150,100);   // Position High
  previousMillis = millis();
}

void robotDance(){
  updateIPK(-20,-20,-150,800);   // Position High

  updateIPK(-20,-20,-200,400);   // Position Low
  updateIPK(-20,-20,-150,800);   // Position High

  updateIPK(-20,-20,-200,100);   // Position Low
  updateIPK(-20,-20,-150,100);   // Position High
}

void checkInactive(){
  if(millis() - previousMillis > MAX_INACTIVE_TIME){
    previousMillis = millis();
    robotDance();
  }
}

int getDistance(){
  // 5v
  float volts = 0;
  // batch of measures
  for (int i=0;i<5;i++){
      volts += analogRead(SENSOR); // value from sensor * (5/254)
      delay(50);
  }
  volts = volts/5; // average
  // int distance = 13*pow(volts, -1); // worked out from datasheet graph
  int distance = volts/5;
  // Serial.println(distance);

  return distance;
}

void goDown(int startFromZ, bool debug = false){
  int previousDistance = 0;
  int distance = getDistance();
  int direction = 1;
  positionZ = startFromZ;
  if(debug == true){
    Serial.println(distance);
  }
  // move until sensors reaches max position (SHARP)
  while (distance >= (previousDistance - 5) && positionZ > -300){
    direction = direction * -1; // invert
    positionZ -= GODOWN_STEP;
    updateIPK(-10*direction,-60,positionZ ,100);
    // delay(200);
    previousDistance = distance;
    distance = getDistance();
    if(debug == true){
      Serial.println(distance);
    }
  }
  positionZ -= 10;
  updateIPK(-20,-40,positionZ ,300);
  updateIPK(20,-40,positionZ ,300);
  updateIPK(-20,-40,positionZ ,300);
}

void loop()
{
  if (Serial.available() > 0) {
    intval = Serial.read();
    val = (char) intval;
    rx_buf[i_rx++] = val;
    if (val == '\n' || val == '\r') {
       Serial.write(rx_buf, i_rx);
       if(strncmp(rx_buf,START_CMD,5) == 0){
         startDelta();
       } else if(strncmp(rx_buf,FAILED_CMD,6) == 0){
         badCode();
       }
       if(strncmp(rx_buf,TEST_CMD,4) == 0){
         Serial.println("test-start");
         // startMillis = millis();
         // int distance;
         // while(millis() - startMillis < 10000){
         //  distance = getDistance();
         //  Serial.println(distance);
         //  delay(300);
         // }
         // Serial.println("test-end");
         goDown(-150, true);
       }
      i_rx = 0;
    }
  }
  checkInactive();
}
